#include "my_me_kernel.h"

#include <stdio.h>
#include <stdlib.h>
#include <cutil.h>
#include <limits.h>

#define MAX_SAD UINT_MAX

sFrameInfos* frameInfos;
sMBInfos* mbTab;

// CPU
unsigned char* refFrame;
unsigned char* curFrame;

__device__ unsigned int SAD( unsigned char* prefFrame, unsigned char* pcurFrame, 
							sFrameInfos* fInfos, 
							unsigned int refmbOffset, unsigned int curmbOffset, 
							int pitch, int mbWidth, int mbHeight, unsigned int lastBestSAD )
{
	unsigned int mySAD = 0;
	int iMb, jMb;
	unsigned char ref, cur;
	
	// hauteur MB
	for( iMb = 0; iMb < mbHeight; iMb++ )
	{		
		// largeur MB
		for( jMb = 0; jMb < mbWidth; jMb++ )
		{
			ref = prefFrame[refmbOffset+iMb*pitch+jMb];
			cur = pcurFrame[curmbOffset+iMb*pitch+jMb];
			
			mySAD = __usad( (unsigned int)cur, (unsigned int)ref, mySAD );
			
			// on arrete la recherche si le SAD actuel est superieur a un SAD plus petit deja trouve
			if( mySAD > lastBestSAD )
			{
				return MAX_SAD;
			}
		}
	}
	return mySAD;
}

__device__ unsigned int bestSADFullSearch( 
								unsigned char* prefFrame, unsigned char* pcurFrame,
								sFrameInfos* fInfos,
								unsigned int mbOffset,
								int pitch,
								int mbWidth, int mbHeight, 
								int MBx, int MBy, 
								int widthInMB, int heightInMB, 
								int xSearchOffset, int ySearchOffset, 
								int* bestSADx, int* bestSADy )
{
	unsigned int myBestSAD = MAX_SAD;
	unsigned int currentSAD = 0;
	unsigned int refmbOffset = mbOffset;
	unsigned int curmbOffset;
	int iSMb, jSMb;
	int xSearchStart, ySearchStart, xSearchEnd, ySearchEnd;	

	// on cale correctement la position et la taille de la fenetre de recherche
	if( MBx == 0 )
	{
		xSearchStart = 0;
		xSearchEnd = xSearchOffset;
	}
	else if( MBx == widthInMB-1 )
	{
		xSearchStart = -xSearchOffset;
		xSearchEnd = 0;
	}
	else
	{
		xSearchStart = -xSearchOffset;
		xSearchEnd = xSearchOffset;
	}
	
	if( MBy == 0 )
	{
		ySearchStart = 0;
		ySearchEnd = ySearchOffset;
	}
	else if( MBy == heightInMB-1 )
	{
		ySearchStart = -ySearchOffset;
		ySearchEnd = 0;
	}
	else
	{
		ySearchStart = -ySearchOffset;
		ySearchEnd = ySearchOffset;
	}

	if( MBx+mbWidth+xSearchEnd > pitch )
	{
		xSearchEnd = pitch-mbWidth;
	}

	if( MBy+mbHeight+ySearchEnd > MBy+fInfos->frameHeight )
	{
		ySearchEnd = fInfos->frameHeight-mbHeight;
	}
	
	// hauteur fenetre de recherche
	for( iSMb = ySearchStart; iSMb < ySearchEnd; iSMb++ )
	{
		curmbOffset = mbOffset+iSMb*pitch;
		// largeur fenetre de recherche
		for( jSMb = xSearchStart; jSMb < xSearchEnd; jSMb++ )
		{
			curmbOffset += jSMb;
			currentSAD = SAD( prefFrame, pcurFrame,fInfos, refmbOffset, curmbOffset, pitch, mbWidth, mbHeight, myBestSAD );
			
			// si on vient de trouver un SAD minimal, on le sauvegarde, ainsi que les coordonnees
			if( currentSAD < myBestSAD )
			{
				myBestSAD = currentSAD;
				*bestSADx = jSMb;
				*bestSADy = iSMb;
			}

			// on ne pourra pas trouver mieux
			if( myBestSAD == 0 )
			{
				return myBestSAD;
			}
		}
	}
	return myBestSAD;
}

__device__ void seachAllPartitionsForOneMB( unsigned char* prefFrame, unsigned char* pcurFrame, 
											sFrameInfos* fInfos, sMBInfos* pcurrMB,
											int widthInMB, int heightInMB,
											int pitch )
{	
	unsigned int mbOffset = pcurrMB->offset;
	int MBx = pcurrMB->uMBx;
	int MBy = pcurrMB->uMBy;

	// 16x16
	pcurrMB->res16x16.SAD = bestSADFullSearch(prefFrame, pcurFrame, fInfos, mbOffset, pitch, 16, 16, MBx, MBy, widthInMB, heightInMB, 16, 16, &pcurrMB->res16x16.MVx, &pcurrMB->res16x16.MVy );
}

__global__ void doME( sFrameInfos* fInfos,unsigned char* prefFrame, unsigned char* pcurFrame, sMBInfos* pmbTab )
{
	const int tid = blockIdx.y*gridDim.x + blockIdx.x;
 
	if( tid < fInfos->uNumMBs )
	{
		seachAllPartitionsForOneMB( prefFrame, pcurFrame, fInfos, &pmbTab[ tid ], 
			fInfos->widthInMBs, fInfos->heightInMBs,
			fInfos->pitchPixels );
	}
}

extern "C" void doME_GPU(sFrameInfos *fInfos)
{
	dim3 dimBlock( 1, 1 ); // 1 thread par block
	dim3 dimGrid( fInfos->widthInMBs, fInfos->heightInMBs ); // la grille
	doME<<< dimGrid, dimBlock >>>( frameInfos, refFrame, curFrame, mbTab );
}

extern "C" void setupGPUFrames( int w, int h )
{
	printf( "Allocating ref and cur frames in GPU\n" );
	int size = w*h*sizeof(unsigned char);
	
    //init de la structure
    CUDA_SAFE_CALL( cudaMalloc( (void**)&frameInfos, sizeof(sFrameInfos) ) );
    
    // Init des Frames
    CUDA_SAFE_CALL( cudaMalloc((void **)&refFrame, size) );
    CUDA_SAFE_CALL( cudaMalloc((void **)&curFrame, size) );
}

extern "C" void copyFramesToGPU( int w, int h, unsigned char* refdata, unsigned char* curdata )
{
	printf( "Copying ref and cur frames to GPU\n" );
	int size = w*h*sizeof(unsigned char);
	
    //copie des images
    CUDA_SAFE_CALL( cudaMemcpy( refFrame, refdata, size, cudaMemcpyHostToDevice ) );
    CUDA_SAFE_CALL( cudaMemcpy( curFrame, curdata, size, cudaMemcpyHostToDevice ) );
}

extern "C" void deleteGPUFrames()
{
	printf( "Deleting ref and cur frames in GPU\n" );
    CUDA_SAFE_CALL( cudaFree( refFrame ) );
    CUDA_SAFE_CALL( cudaFree( curFrame ) );
}

extern "C" void copyFrameInfosToGPU( sFrameInfos* fInfos, sMBInfos* mbTab_cpu )
{
	printf( "Copying frameInfos to GPU\n" );

    //copie de structure
    CUDA_SAFE_CALL( cudaMemcpy( frameInfos, fInfos, sizeof(sFrameInfos), cudaMemcpyHostToDevice ) );
	
    // vecteurs SMInfos malloc
	CUDA_SAFE_CALL( cudaMalloc( (void**)&mbTab, fInfos->uNumMBs*sizeof(sMBInfos) ) );

    // vecteurs SMInfos
    CUDA_SAFE_CALL( cudaMemcpy( mbTab, mbTab_cpu, fInfos->uNumMBs*sizeof(sMBInfos), cudaMemcpyHostToDevice ) );
}

extern "C" void getFrameInfosFromGPU( sFrameInfos* fInfos ,sMBInfos* mbTab_cpu)
{
	printf( "Getting frameInfos from GPU\n" );
	
	CUDA_SAFE_CALL( cudaMemcpy( mbTab_cpu, mbTab, fInfos->uNumMBs*sizeof(sMBInfos), cudaMemcpyDeviceToHost ) );
	
	CUDA_SAFE_CALL( cudaFree( mbTab ) );
}
